"""
"""

from PyQt4.QtCore  import *
from PyQt4.QtGui  import *

import re

import Connection
import Constants
import Dock.Plotting
import Dock.LinearAlgebra
import Dock.Transforms
import Dock.Cosmology
import Dock.GeneralFunctions
import Dock.Statistics
import Dock.StellarDynamics
import Preferences
import RemoteFile


class Actions:
	"""
	Overview
	========
	This class holds virtually all of the QAction's that the client will perform.  That is not to say it holds
	all of the functionality, but rather the middleman to the functionality.  These actions define what happens
	when they are activated, and then other things like buttons use a QAction to actually do something useful.

	Every action in the menu bar is contained here.  They exist here and not in the menu class because other
	things, such as the tool bar also use them.  They are grouped according to package/section in a python list.
	This is so that other objects can unroll the entire list and get all of the actions at once.
	
	If you are adding an action to an existing package, then its as easy as copying a previous entry.  It will be
	propogated through the entire program from here.

	Packages are a tad trickier to implement, but not that much. You can mimic the way they are laid out here,
	however, objects like the tool/menu bar will need to be aware the new package exists.  See the individual
	object's class for details.
	"""

	def __init__ (self, bridge):
		"""
		Overview
		========
		This initialization method takes only the bridge to
		set up its interconnectivity.  Other than that it is
		simply setting up lists of QActions and Actions.

		QActions are used for non-dock related actions.  This
		is for things like open/save which are independant.

		Actions are used for dock related actions.  They handle
		some of the hiding/showing of the widget as well as
		enabling/disabling the action when its in use.

		It looks painful looking at all of the definitions in here,
		however they get propogated everywhere else, so a little
		pain here helps the other widgets that need these actions
		defined.

		@param bridge: The bridge object which connects all interactive components together.
		@type bridge: L{Bridge<client.Bridge.Bridge>}
		"""

		self._bridge = bridge
		bridge.setActions (self)

		# When defining an action, actions which interact with the dock
		# (ones whose actions become temporarily disabled), should be
		# using the class Action, rather than QAction. An example of this
		# would be any of the transformation functions.  Actions, such as
		# connect do not use the dock, and so should be initialized as a
		# QAction.


		#******************************************
		#File Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._file = []

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/connect.png"), "Connect", None))
		self._file[-1].setStatusTip ("Allows you to connect to your cluster")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.connect)
		self._file[-1].setShortcut ("Ctrl+c")

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/disconnect.png"), "Disconnect", None))
		self._file[-1].setStatusTip ("Disconnects you from the cluster")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.disconnect)
		self._file[-1].setShortcut ("Ctrl+d")

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/update.png"), "Update", None))
		self._file[-1].setStatusTip ("Updates the interface for all changes that may have occurred")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self._bridge.Update)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/open.png"), "Open", None))
		self._file[-1].setStatusTip ("Open a data data file to a region")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.importFile)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/save.png"), "Save", None))
		self._file[-1].setEnabled (False)
		self._file[-1].setStatusTip ("Save a region to a file")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.exportFile)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/python.png"), "Save IPython Log", None))
		self._file[-1].setStatusTip ("Save your IPython's session log to Html or Text")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self._bridge.SaveIPythonSession)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/savebatch.png"), "Save Batch File", None))
		self._file[-1].setStatusTip ("Save commands you've issued to a batch file to redo later")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.saveBatch)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/loadbatch.png"), "Load Batch File", None))
		self._file[-1].setStatusTip ("Loads a batch file and executes all of its commands")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.loadBatch)

		self._file.append (QAction (QIcon (Constants.ICON_DIR + "/file/preferences.png"), "Preferences", None))
		self._file[-1].setStatusTip ("Brings up dialog for changing basin_client settings")
		QObject.connect (self._file[-1], SIGNAL ("triggered ()"), self.preferences)


		#******************************************
		#General Functions Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._genf = []

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/random.png"), \
			QString ("Random Attribute"), bridge))
		self._genf[-1].setStatusTip ("Generates a random atribute of any given size")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.AttributeRandom)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/constant.png"), \
			QString ("Constant Attribute"), bridge))
		self._genf[-1].setStatusTip ("Generates an atribute, filled with a constant, of any given size")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.AttributeConstant)

		self._genf.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Attribute Index"), bridge))
		self._genf[-1].setStatusTip ("Generates an attribute of indicies into your input data")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.AttributeIndex)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/real.png"), \
			"Real", bridge))
		self._genf[-1].setStatusTip ("Retrieves an attribute's real part")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Real)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/imaginary.png"), \
			"Imaginary", bridge))
		self._genf[-1].setStatusTip ("Retrieves an attribute's imaginary part")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Imaginary)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/sort.png"), "Sort", bridge))
		self._genf[-1].setStatusTip ("Sorts an attribute")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Sort)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/sortindex.png"), \
			"Index Sort", bridge))
		self._genf[-1].setStatusTip ("Sorts an attribute by index")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.IndexSort)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/absval.png"), \
			"Absolute Value", bridge))
		self._genf[-1].setStatusTip ("Takes the absolute value of an attribute")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.AbsoluteValue)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/pow.png"), \
			"Power", bridge))
		self._genf[-1].setStatusTip ("Takes an attribute to the nth power")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Pow)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/squareroot.png"), \
			QString ("Square Root"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's square root")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.SquareRoot)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/log.png"), \
			QString ("log"), bridge))
		self._genf[-1].setStatusTip ("Takes the natural log of the attribute")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Log)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/log10.png"), \
			QString ("log10"), bridge))
		self._genf[-1].setStatusTip ("Takes log base 10 of an attribute")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Log10)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/exp.png"), \
			QString ("exp"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute as the power of e")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Exp)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/exp10.png"), \
			QString ("exp10"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute as the power of 10")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Exp10)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/sin.png"), \
			QString ("sin"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's sin")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Sin)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/cos.png"), \
			QString ("cos"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's cos")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Cos)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/cosh.png"), \
			QString ("cosh"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's cosh")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Cosh)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/sinh.png"), \
			QString ("sinh"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's sinh")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Sinh)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/arccos.png"), \
			QString ("acos"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's acos")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Acos)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/arcsin.png"), \
			QString ("asin"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's asin")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Asin)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/acosh.png"), \
			QString ("acosh"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's acosh")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Acosh)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/asinh.png"), \
			QString ("asinh"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's asinh")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Asinh)

		self._genf.append (Action (QIcon (Constants.ICON_DIR + "/general_functions/atan.png"), \
			QString ("atan"), bridge))
		self._genf[-1].setStatusTip ("Takes an attribute's atan")
		QObject.connect (self._genf[-1], SIGNAL ("triggered ()"), self._bridge._dock.Atan)


		#******************************************
		#Plotting Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._plot = []

		self._plot.append (Action (QIcon (Constants.ICON_DIR + "/plotting/plot.png"), \
			QString ("Gnuplot"), bridge))
		self._plot[-1].setStatusTip ("Shows the dialog for plotting in Gnuplot")
		QObject.connect (self._plot[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.Gnuplot)

		self._plot.append (Action (QIcon (Constants.ICON_DIR + "/plotting/visit.png"), \
			QString ("VisIt Plot"), bridge))
		self._plot[-1].setStatusTip ("Shows the dialog for plotting in VisIt")
		QObject.connect (self._plot[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.VisIt)

		self._plot.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("PyLab"), bridge))
		self._plot[-1].setStatusTip ("Shows the dialog for plotting in PyLab")
		QObject.connect (self._plot[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.PyLab)


		#******************************************
		#Cosmology Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._cosm = []

		self._cosm.append (Action (QIcon (Constants.ICON_DIR + "/cosmology/rdist.png"), \
			QString ("Redshift To Distance"), bridge))
		self._cosm[-1].setStatusTip ("Shows the dialog for " + \
			"performing Redshift To Distance")
		QObject.connect (self._cosm[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.RedshiftToDistance)

		self._cosm.append (Action (QIcon (Constants.ICON_DIR + "/cosmology/rgrowth.png"), \
			QString ("Redshift To Growth"), bridge))
		self._cosm[-1].setStatusTip ("Shows the dialog for performing Redshift To Growth")
		QObject.connect (self._cosm[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.RedshiftToGrowth)

		self._cosm.append (Action (QIcon (Constants.ICON_DIR + "/cosmology/hubble.png"), \
			QString ("Hubble Constant"), bridge))
		self._cosm[-1].setStatusTip ("Shows the dialog for getting the Hubble Constant")
		QObject.connect (self._cosm[-1], SIGNAL ("triggered ()"), self._bridge._dock.Hubble)


		#******************************************
		#Stellar Dynamics Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._stel = []

		self._stel.append (Action (QIcon (Constants.ICON_DIR + "/stellar_dynamics/centerofmass.png"), \
			QString ("Center Of Mass"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting a center of mass")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.CenterOfMass)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Local Density"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting the local density")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.LocalDensity)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Density List"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting the density list")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.DensityList)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Density Center"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting the density's center")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.DensityCenter)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Radial Profile"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting a radial profile")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.RadialProfile)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Density Profile"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting a density profile")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.DensityProfile)

		self._stel.append (Action (QIcon (Constants.ICON_DEFAULT), \
			QString ("Disp Profile"), bridge))
		self._stel[-1].setStatusTip ("Shows the dialog for getting a disp profile")
		QObject.connect (self._stel[-1], SIGNAL ("triggered ()"), self._bridge._dock.DispProfile)



		#******************************************
		#Statistics Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._stat = []

		self._stat.append (Action (QIcon (Constants.ICON_DIR + "/statistics/mean.png"), \
			QString ("Mean"), bridge))
		self._stat[-1].setStatusTip ("Shows the dialog for getting the mean")
		QObject.connect (self._stat[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.Mean)

		self._stat.append (Action (QIcon (Constants.ICON_DIR + "/statistics/stddev.png"), \
			QString ("Standard Deviation"), bridge))
		self._stat[-1].setStatusTip ("Shows the dialog for getting " + \
			"the standard deviation")
		QObject.connect (self._stat[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.StandardDeviation)

		self._stat.append (Action (QIcon (Constants.ICON_DIR + "/statistics/histogram.png"), \
			QString ("Histogram"), bridge))
		self._stat[-1].setStatusTip ("Shows the dialog for getting " + \
			"a histogram's data")
		QObject.connect (self._stat[-1], SIGNAL ("triggered ()"), \
			self._bridge._dock.Histogram)


		#******************************************
		#Transformations Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._tran = []

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/etoc.png"), \
			QString ("Equatorial To Cartesian"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an equatorial to cartesian transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.EquatorialToXyz)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/ctoe.png"), \
			QString ("Cartesian To Equatorial"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an cartesian to equatorial transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.XyzToEquatorial)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/ptoc.png"), \
			QString ("Polar To Cartesian"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an polar to cartesian transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.PolarToXy)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/ctop.png"), \
			QString ("Cartesian To Polar"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an cartesian to polar transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.XyToPolar)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/stoc.png"), \
			QString ("Spherical To Cartesian"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an spherical to cartesian transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.SphericalToXyz)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/ctos.png"), \
			QString ("Cartesian To Spherical"), bridge))
		self._tran[-1].setStatusTip ("Shows the idialog for an cartesian to spherical transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.XyzToSpherical)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/cytoc.png"), \
			QString ("Cylindrical To Cartesian"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an cylindrical to cartesian transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.CylindricalToXyz)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/ctocy.png"), \
			QString ("Cartesian To Cylindrical"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an cartesian to cylindrical transformation")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.XyzToCylindrical)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/fft.png"), \
			QString ("FFT"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for FFT transformations")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.Fft)

		self._tran.append (Action (QIcon (Constants.ICON_DIR + "/transformations/convolve.png"), \
			QString ("Convolve"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for an convolving")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.Convolve)

		self._tran.append (Action (QIcon (QString (Constants.ICON_DIR + "/transformations/compdens.png")), \
			QString ("Compute Density"), bridge))
		self._tran[-1].setStatusTip ("Shows the dialog for computing density")
		QObject.connect (self._tran[-1], SIGNAL ("triggered ()"), self._bridge._dock.ComputeDensity)


		#******************************************
		#Linear Algebra Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._linalg = []

		self._linalg.append (Action (QIcon (Constants.ICON_DIR + "/linear_algebra/matrixcreate.png"), \
			"Create Matrix", bridge))
		self._linalg[-1].setStatusTip ("Shows the dialog for creating a matrix")
		QObject.connect (self._linalg[-1], SIGNAL ("triggered ()"), self._bridge._dock.MatrixCreate)

		self._linalg.append (Action (QIcon (Constants.ICON_DIR + "/linear_algebra/identitymatrixcreate.png"), \
			"Create Identity Matrix", bridge))
		self._linalg[-1].setStatusTip ("Shows the dialog for creating an identity matrix")
		QObject.connect (self._linalg[-1], SIGNAL ("triggered ()"), self._bridge._dock.MatrixCreateIdentity)

		self._linalg.append (Action (QIcon (Constants.ICON_DIR + "/linear_algebra/matrixinvert.png"), \
			"Invert Matrix", bridge))
		self._linalg[-1].setStatusTip ("Shows the dialog for inverting a matrix")
		QObject.connect (self._linalg[-1], SIGNAL ("triggered ()"), self._bridge._dock.MatrixInvert)

		self._linalg.append (Action (QIcon (Constants.ICON_DIR + "/linear_algebra/matrixmult.png"), \
			"Multiply Matrix", bridge))
		self._linalg[-1].setStatusTip ("Shows the dialog for multilying matrices")
		QObject.connect (self._linalg[-1], SIGNAL ("triggered ()"), self._bridge._dock.MatrixMult)


		#******************************************
		#Help Actions
		#
		#  define them here and they will propagate
		#  themselves throughout menus/toolbars
		#******************************************
		self._help = []

		self._help.append (QAction (QIcon (Constants.ICON_DIR + "/basin.png"), "About", None))
		self._help[-1].setStatusTip ("Provides general information regarding BASIN")
		QObject.connect (self._help[-1], SIGNAL ("triggered ()"), self.helpAbout)

		self._help.append (QAction (QIcon (Constants.ICON_DIR + "/help/wiki.png"), "Wiki", None))
		self._help[-1].setStatusTip ("Provides URL's for BASIN's wiki")
		QObject.connect (self._help[-1], SIGNAL ("triggered ()"), self.helpWiki)

	def menuConnect (self):
		"""
		Overview
		========
		This method will enable or disable actions depending on whether they should be used if you are
		connected.  This, of course, gets called whenever your connectivity changes.  You can, though
		you shouldn't, add any dock-related widgets to this method.
		"""

		for act in self._file:
			if   str (act.text ()) == "Connect":
				act.setEnabled (False)
			elif str (act.text ()) == "Disconnect":
				act.setEnabled (True)
			elif str (act.text ()) == "Open":
				act.setEnabled (True)
			elif str (act.text ()) == "Save IPython Log":
				act.setEnabled (True)
			elif str (act.text ()) == "Update":
				act.setEnabled (True)
			elif str (act.text ()) == "Save Batch File":
				act.setEnabled (True)
			elif str (act.text ()) == "Load Batch File":
				act.setEnabled (True)

	def menuDisconnect (self):
		"""
		Overview
		========
		This argument-less method will enable/disable
		actions depending on whether they should be used
		if you are disconnected.  This, of course, gets
		called whenever your connectivity changes.  You
		can, though you shouldn't, add any dock-related
		widgets to this method.
		"""

		for act in self._file:
			if   str (act.text ()) == "Connect":
				act.setEnabled (True)
			elif str (act.text ()) == "Disconnect":
				act.setEnabled (False)
			elif str (act.text ()) == "Open":
				act.setEnabled (False)
			elif str (act.text ()) == "Save IPython Log":
				act.setEnabled (False)
			elif str (act.text ()) == "Update":
				act.setEnabled (False)
			elif str (act.text ()) == "Save Batch File":
				act.setEnabled (False)
			elif str (act.text ()) == "Load Batch File":
				act.setEnabled (False)

	def menuRegionSelected (self):
		"""
		Overview
		========
		Assuming everything is connected and parallel, this will enable the Save Action.  Presumably
		this is only called when you click on a Region in the RegionViewer.
		"""

		if self._bridge._connected and self._bridge._parallel:
			for act in self._file:
				if   str (act.text ()) == "Save":
					act.setEnabled (True)

	def menuRegionNotSelected (self):
		"""
		Overview
		========
		Assuming everything is connected and parallel, this will disable the Save Action.  Presumably
		this is only called when you click on a non-Region in the RegionViewer.
		"""

		for act in self._file:
			if   str (act.text ()) == "Save":
				act.setEnabled (False)

	def importFile (self):
		"""
		Overview
		========
		This will create an instance of the RemoteFileBrowser. After it is done, it will retrieve the
		finalresult from the RemoteFileBrowser, and if one actually exists it will create an ipython
		statement that will go ahead and open up the file.
		
		This is where the naming convention of reg# is generated. It does this by getting a list of all
		the top-level regions and it starts counting from reg0 up, and as soon as it finds a name not
		being used, it takes it.
	
		One should always remember that the filename retrieved will be one on the remote host, and that
		the Region constructor in basin_remote handles which BASIN_IO sub-class to use to open the file,
		so this doesn't need to.
		"""

		# Attempt a filename from the user, and if successful
		# go ahead and open the file.
		dialog = RemoteFile.RemoteFileBrowser (self._bridge)
		for file in dialog.finalresult:
			# Finds a valid name that isn't already in use
			topRegions = self._bridge._ipyShell.getTopLevelRegions ()
			count = 0
			var = "reg" + str (count)
			while topRegions.count (var) > 0:
				count += 1
				var = "reg" + str (count)

			# execute the Region constructor in ipython
			self._bridge.CursorBusy ()
			self._bridge._ipyShell.Command (str (var) + " = Region (\"" + str (file) + "\")")
			self._bridge.CursorNorm ()
				

	def exportFile (self):
		"""
		Overview
		========
		This will create an instance of the RemoteFileBrowserExport. After it is done, it will retrieve
		the finalresult from the RemoteFileBrowser, and if one actually exists it will create an ipython
		statement that will go ahead and save up the file.
	
		Unlike opening files, saving them requires knowledge as to what format you want.  It is here
		that you will need to decide what format to use.  I use regular expressions to determine if
		the filename chosen works with our formats to figure out what to save it as.
	
		Note that there is already a mechanism to stop users from opening this until they have selected
		a Region, however, this will also safeguard against users from saving non-regions.
		"""

		region = self._bridge._regionViewer.getSingleSelectedItem ()
		if region != None and region.Type () == "REGION":
			dialog = RemoteFile.RemoteFileBrowserExport (self._bridge)
			file = dialog.finalresult[0]
			if re.compile (".*\.h5").match (str (file)):
				self._bridge.CursorBusy ()
				self._bridge._ipyShell.Command (str (region._shellAccess) + \
					".SaveHDF5 (\"" + file + "\")")
				self._bridge.CursorNorm ()
			elif re.compile (".*\.(dat|txt)").match (str (file)):
				self._bridge.CursorBusy ()
				self._bridge._ipyShell.Command (str (region._shellAccess) + \
					".SaveTEXT (\"" + file + "\")")
				self._bridge.CursorNorm ()
		else:
			self._bridge._statusBar.showTimed ("You may only export Regions!", STATUS_WAIT)

	def saveBatch (self):
		"""
		Overview
		========
		Saving a batch is a mildy complicated process because the idea is to save the commands from
		the entire session, not just while a single user was connected to a connection. This makes it
		so that whenever any command from the ipython shell is entered, it is likely also sent to an
		ipython server script that saves it to the server's drive.  Thus, to save a batch, this server
		buffer file must be read, which is what is done here.

		After the server's buffer history is read, it is stored and sent to the BatchSave class which
		will display a dialog which aids the user in selecting/deselecting which commands to keep and
		which to throw out.   The BatchSave class also handles the actual file io of saving the newly
		created batch file.
		
		Please note that a batch file is nothing more than a clear text file of ipython commands, except
		the first line is a number, the number of lines to read.  Also note that like any interactive
		python session, when you enter multiple line constructs, an extra newline is reuqired to end the
		construct, and this line counts towards the top number.
		"""

		# Make sure you aren't in the middle of a multi-line construct, and if you
		# are then end it by sending a newline...
		if len (self._bridge._ipyShell._commandLine._multiLineHistoryBuffer) > 0:
			self._bridge._ipyShell.Command ("")
		# Read in the server buffer history
		history = self._bridge._ipyShell.ControllerCommand ("buffer_history.read ()")
		if history.find ("rror") != -1:
			self._bridge.Error ("BASIN_actions.py", \
				"BasinActions", "saveBatch", history)
		# lop off the first 4 garbage lines from ipython
		# Note that I haven't extensivley tested this with the preferences
		# option that reduces output.
		history = history.split ("\n")[4:]
		if history == []: return

		BatchSave (self._bridge, history)

	def loadBatch (self):
		"""
		Overview
		========
		This method will use the QFileDialog widget from Qt to open up the batch file.  Once a file has
		been chosen, it will read the first line, the total number of lines to process. Originally this
		was to make a progresbar, which calculates its percentage from number completed out of total
		things. I never got around to doing this, so if you're interested look up the class QProgressBar.

		After that, it reads every line thereafter and executes it. This was written before my
		IpythonView's BatchCommands was written, so this could likely be written slightly more concisely
		with that, but this should suffice either way.
		"""

		dialog = QFileDialog (None, QString ("Load Script"),
			QString (str (QDir.home ().path ())),
			QString ("BASIN Script (*.ipy)"))
		dialog.setWindowIcon (QIcon (QString (Constants.ICON_DIR + "/file/loadbatch.png")))
		dialog.setViewMode (QFileDialog.Detail)
		dialog.setFileMode (QFileDialog.ExistingFile)
		dialog.setAcceptMode (QFileDialog.AcceptOpen)
		dialog.exec_ ()

		files = dialog.selectedFiles ()
		if not files.isEmpty () and dialog.result () == 1:
			file = QFile (str (files[0]))
			file.open (QFile.ReadOnly)
			stream = QTextStream (file)

			commands = []
			while stream.atEnd () == False:
				commands.append (str (stream.readLine ()))
			file.close ()
			self._bridge._ipyShell.BatchCommands (commands)

	def helpWiki (self):
		"""
		Overview
		========
		This is just a quick QMessageBox that pops up and shows the user the clickable urls for our
		wiki-documentation sites.
		"""

		dialog = QMessageBox (QMessageBox.NoIcon, "BASIN Wiki", 
		"<a href='http://www.physics.drexel.edu/basin_trac/kernel_trac'>Kernel Trac</a><br><br>" +
		"<a href='http://www.physics.drexel.edu/basin_trac/remote_trac'>Remote Trac</a><br><br>" +
		"<a href='http://www.physics.drexel.edu/basin_trac/client_trac'>Client Trac</a>")
		dialog.setIconPixmap (QPixmap (Constants.ICON_DIR + "/help/wiki_large.png"))
		dialog.setWindowIcon (QIcon (Constants.ICON_DIR + "/help/wiki.png"))
		dialog.exec_ ()


	def helpAbout (self):
		"""
		Overview
		========
		This is just a quick QMessageBox that pops up and shows the user a brief synopsis of Basin.
		"""

		QMessageBox.information (self._bridge._mainWindow, "About BASIN",
		"Beowulf PC clusters are now widely used in many branches  of\n" +
		"science where they have become key tools for the analysis of large\n" +
		"datasets resulting from laboratory experiments, field observations\n" +
		"or numerical simulations. However, although Beowulf clusters have\n" +
		"been of great benefit to scientific research, no publicly available\n" +
		"parallelized software package exist to perform  even the most\n" +
		"common operations needed in the analysis and visualization of\n" +
		"large datasets. In sharp contrast with the situation for\n" +
		"non-parallelized data analysis, where a wealth of software packages\n" +
		"performing standard data-reduction tasks are available to\n" +
		"researchers (e.g. IRAF, MIDAS), for parallel data analysis and\n" +
		"visualization each individual group has tended to develop its own\n" +
		"set of specialized tools. The increasing diffusion  of multi-processor\n" +
		"PCs further increase the demand of software packages for parallel\n" +
		"data analysis.\n\n" +
		"BASIN  is a  flexible, integrated suite of tools for parallel data\n" +
		"analysis and visualization which has been designed and is currently\n" +
		"been developed to address these concerns; BASIN is aimed at\n" +
		"providing researchers with a  package which allows them to harness\n" +
		"the power of Beowulf PC clusters and multi-processor  machines\n" +
		"without necessarily being experts in parallel programming. While\n" +
		"BASIN is motivated by problems particular to astrophysics, its core\n" +
		"functionalities and parallel data distribution and manipulation are\n" +
		"general end provide a number of general  to easily develop libraries\n" +
		"specific to data analysis in a number of different fields.")

	def connect (self):
		"""
		Overview
		========
		All of the functionality is handled within the ConnectionDialog, so that should be your primary
		source of information.

		Roughly speaking, it brings up a dialog which manages different profiles for connection to ipython
		sessions.  These profiles get saved to ~/.basin/client/connection.conf to be loaded later on.
		"""

		dialog = Connection.ConnectionDialog (self._bridge)

	def disconnect (self):
		"""
		Overview
		========
		Disconnecting is quite easy, since all of the information about your connection is stored locally,
		we can just destroy the rc variable in the ipython session.  This will render you disconnected
		without any harm.
		"""

		self._bridge.makeLocal ()
		self._bridge._ipyShell.SilentCommand ("del (rc)")
		self._bridge.Disconnect ()

	def preferences (self):
		"""
		Overview
		========
		This method will create an intance of the L{Preferences dialog<client.Preferences.Preferences>}.
		The dialog will allow the user to change settings regarding:
		 - General/Miscellaneous
		 - Connection
		 - Toolbar
		 - Appearance
		"""

		Preferences.Preferences (self._bridge)


class BatchSave (QDialog):
	def __init__ (self, bridge, history):
		QDialog.__init__ (self, bridge._mainWindow)

		self._bridge = bridge
		self._history = history

		mainRect = self._bridge._mainWindow.geometry  ()
		width = mainRect.width ()
		height = mainRect.height ()
		x = mainRect.x () + int (width * 0.10)
		y = mainRect.y () + int (height * 0.25)
		width = int (width * 0.8)
		height = int (height * 0.5)

		dialog = QDialog ()
		dialog.setGeometry (x, y, width, height)
		dialog.setWindowIcon (QIcon (Constants.ICON_DIR + "/file/savebatch.png"))
		dialog.setWindowTitle ("Save Script")
		layout = QVBoxLayout (dialog)
		dialog.setLayout (layout)
		mainlayout = QHBoxLayout ()
		layout.addLayout (mainlayout)

		gbox = QGroupBox ("Users")
		mainlayout.addWidget (gbox)
		gboxlayout = QVBoxLayout ()
		gbox.setLayout (gboxlayout)

		vlayout = QVBoxLayout ()
		mainlayout.addLayout (vlayout)
		self._table = QTableWidget (len (self._history), 2, None)
		self._table.setShowGrid (False)
		QObject.connect (self._table, SIGNAL ("cellClicked (int,int)"), self.toggleCell)
		vlayout.addWidget (self._table)
		headers = QStringList (["Command", "User"])
		self._table.setHorizontalHeaderLabels (headers)

		total = len (self._history)
		count = 0
		self.users = []
		# self._history is a list of strings of the format:
		#    user@command
		_flip = False
		for each in self._history:
			at = each.find ("@")
			userName = each[:at]
			if userName[0] == "#":
				userName = userName[1:]
				_flip = True
			if self.users.count (userName) == 0:
				self.users.append (userName)
			user = QTableWidgetItem (userName)
			user.setFlags (Qt.ItemIsEnabled)
			command = QTableWidgetItem (str (each[at + 1:]))
			command.setFlags (Qt.ItemIsEditable | Qt.ItemIsEnabled)
			if _flip == True:
				color = QColor (80, 80, 80)
				command.setBackgroundColor (color)
				user.setBackgroundColor (color)
				_flip = False
			self._table.setItem (count, 0, command)
			self._table.setItem (count, 1, user)
			count += 1
		self._table.resizeColumnsToContents ()

		gboxlayout.addWidget (QLabel (QString (
			"You may (de)select\n" +
			"entire groups of\n" +
			"commands to save to\n" +
			"the batch based on\n" +
			"the user who entered\n" +
			"the command.")))
		for each in self.users:
			BatchCheck (each, self, gboxlayout)

		buttonLayout = QHBoxLayout ()
		vlayout.addLayout (buttonLayout)
		saveButton = Constants.BQPushButton ("Save", buttonLayout)
		QObject.connect (saveButton, SIGNAL ("clicked ()"), dialog.accept)
		cancelButton = Constants.BQPushButton ("Cancel", buttonLayout)
		QObject.connect (cancelButton, SIGNAL ("clicked ()"), dialog.reject)

		layout.addWidget (QLabel (QString (
			"Clicking on a row will either select it (turning it white), or deselect\n" + \
			"it (turning it gray). Only those selected (white) commands will be saved\n" + \
			"to a batch file.")))

		self._rows = self._table.rowCount () 
		dialog.exec_ ()

		if dialog.result () == 1: # 'Save'
			self.save ()

	def toggleCell (self, row, col):
		if self._table.item (row, col).backgroundColor () == QColor (255, 255, 255):
			color = QColor (80, 80, 80)
			self._table.item (row, 0).setBackgroundColor (color)
			self._table.item (row, 1).setBackgroundColor (color)
		else:
			color = QColor (255, 255, 255)
			self._table.item (row, 0).setBackgroundColor (color)
			self._table.item (row, 1).setBackgroundColor (color)

	def selectUser (self, user):
		color = QColor (255, 255, 255)
		for index in xrange (self._table.rowCount ()):
			if str (self._table.item (index, 1).text ()) == str (user):
				self._table.item (index, 0).setBackgroundColor (color)
				self._table.item (index, 1).setBackgroundColor (color)

	def deselectUser (self, user):
		color = QColor (80, 80, 80)
		for index in xrange (self._table.rowCount ()):
			if str (self._table.item (index, 1).text ()) == str (user):
				self._table.item (index, 0).setBackgroundColor (color)
				self._table.item (index, 1).setBackgroundColor (color)

	def getSelectedCommands (self):
		commands = []
		for index in xrange (self._table.rowCount ()):
			if self._table.item (index, 0).backgroundColor () == QColor (255, 255, 255):
				commands.append (str (self._table.item (index, 0).text ()))
		return commands

	def ReWriteBuffer (self):
		self._bridge._ipyShell.ControllerCommand ("buffer_history.delete ()")
		for index in xrange (self._table.rowCount ()):
			if self._table.item (index, 0).backgroundColor () == QColor (255, 255, 255):
				self._bridge._ipyShell.ControllerCommand ("buffer_history.append (\"" + \
					str (self._table.item (index, 1).text ()) + "@" + \
					str (self._table.item (index, 0).text ()).replace ("\"", "\\\\\"") + "\")")
			else:
				self._bridge._ipyShell.ControllerCommand ("buffer_history.append (\"#" + \
					str (self._table.item (index, 1).text ()) + "@" + \
					str (self._table.item (index, 0).text ()).replace ("\"", "\\\\\"") + "\")")

	def save (self):
		self.ReWriteBuffer ()
		dialog = QFileDialog (None, QString ("Save Script"),
			QString (str (QDir.home ().path ())),
			QString ("BASIN Script (*.ipy)"))
		dialog.setWindowIcon (QIcon (Constants.ICON_DIR + "/file/savebatch.png"))
		dialog.setViewMode (QFileDialog.Detail)
		dialog.setFileMode (QFileDialog.AnyFile)
		dialog.setAcceptMode (QFileDialog.AcceptSave)
		dialog.exec_ ()

		files = dialog.selectedFiles ()
		if not files.isEmpty ():
			fileName = files[0]
			if fileName[-4:] != ".ipy":
				fileName = fileName + ".ipy"

			file = QFile (fileName)
			file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | \
				QFile.WriteUser | QFile.ReadGroup | QFile.WriteGroup | QFile.ReadOther)
			file.open (QIODevice.WriteOnly)
			commands = self.getSelectedCommands ()
			for command in commands:
				file.write (str (command) + "\n")
			file.close ()

class BatchCheck (QCheckBox):
	def __init__ (self, label, parent, layout):
		QCheckBox.__init__ (self, label)
		QObject.connect (self, SIGNAL ("stateChanged (int)"), self.toggleSelection)
		self._parent = parent
		self._name = label
		self.setCheckState (Qt.Checked)

		layout.addWidget (self)

	def toggleSelection (self, val):
		if val == 2:
			self._parent.selectUser (self._name)
		elif val == 0:
			self._parent.deselectUser (self._name)




class Action (QAction):
	"""
	Overview
	========
	Action reimplements QAction for the purpose of L{DockWidget<client.Dock.Widget>}s. Actions will automatically have their
	triggers connected to their corresponding DockWidget
	"""

	def __init__ (self, icon, name, bridge, object = None):
		self._bridge = bridge
		QAction.__init__ (self, icon, name, None)
		QObject.connect (self, SIGNAL ("triggered ()"), self.toggleEnabled)
		QObject.connect (self, SIGNAL ("triggered ()"), self.ShowWidget)
		self._widget = None
		if object:
			self._widget = object (bridge)

	def toggleEnabled (self):
		self._bridge._dock.DisableAction (self)

	def ShowWidget (self):
		if self._widget:
			if self._bridge._dock._current != self._widget:
				self._bridge._dock._current = self._widget
				self._widget.setVisible (True)
				try:
					self._widget.guessDefaults ()
				except AttributeError:
					pass
				try:
					self._widget.DefaultCheck ()
				except AttributeError:
					pass

