"""
Overview
========
This module provides the specific Dock Widgets used for functions in the Basin Remote's I{Transformation} package.
These Dock Widget implementations provide an interface for selecting input data and designating output for the
individual functions.
"""

import PyQt4.QtCore
import PyQt4.QtGui

import Constants
import DataGuesser
import Widget
import Holders

class AbstractTransform (Widget.DockWidget):
	"""
	Overview
	========
	This class is essentially a virtual class, in that it really should never be instantiated, but rather
	further derived into more useful classes.  The way in which it works is by creating a generic template
	that works for virtually all of the transformation fucntions.  It does require that the derived versions
	of this class setup a number of member variables before they use the constructor of AbstractTransform.
	"""

	def __init__ (self, parent, bridge):
		"""
		Overview
		========
		This constructor simply sets the type of the package to "Transformation", after which it
		runs the L{DockWidget<client.Dock.Widget.DockWidget>}'s constructor.
		"""

		self._type = "Transformation"
		Widget.DockWidget.__init__ (self, parent, bridge)

	def _addOptions (self):
		"""
		Overview
		========
		The standard option for an Transform Dock Widget is the "Transform in Place" check box
		which allows users to designate where their output goes.  When it is checked, the output
		will be directed to the List from which the input came from, otherwise they will be
		free variables.
		"""

		box = PyQt4.QtGui.QGroupBox ("Options")
		layout = PyQt4.QtGui.QVBoxLayout ()
		box.setLayout (layout)

		self.pref = PyQt4.QtGui.QCheckBox ("Transform in Place")
		layout.addWidget (self.pref)
		PyQt4.QtCore.QObject.connect (self.pref, PyQt4.QtCore.SIGNAL ("stateChanged (int)"), self.setOutput)
		return box

	def _addCommands (self):
		"""
		Overview
		========
		The stanard commands for the Transform Dock Widget are "Transform" and "Clear".  "Transform"
		will actually execute the transformation if possible, and clear will remove any input data
		the user has already placed in the data holders.
		"""

		box = PyQt4.QtGui.QGroupBox ("Commands")
		layout = PyQt4.QtGui.QHBoxLayout (box)

		transformButton = Constants.BQPushButton ("Transform", layout)
		transformButton.setDefault (True)
		PyQt4.QtCore.QObject.connect (transformButton, PyQt4.QtCore.SIGNAL ("clicked ()"), self.transform)

		clearButton = Constants.BQPushButton ("Clear", layout)
		PyQt4.QtCore.QObject.connect (clearButton, PyQt4.QtCore.SIGNAL ("clicked ()"), self.clear)

		return box

	def _addDataSelection (self):
		"""
		"""

		box = PyQt4.QtGui.QGroupBox ("Data Selection")
		inputDict = {}
		for key in self._inList:
			inputDict[key] = [["ATTRIBUTE"], ["_basin.Attribute"]]
		self.list = Holders.GenericGroupHolder (self._bridge, [["LIST"], ["_basin.List"]], "List", inputDict, self._inList, self._outList)
		self.list.setEnabled (False)
		layout = self.list._layout
		box.setLayout (layout)
		return box

	def clear (self):
		"""
		Overview
		========
		This method will clear all user modifiable fields. For input fields this means returning
		them to None, while for Output fields it returns them to their default values.  This is
		called on successful transform, as well as when the user clicks clear.  Finally, it is also
		called before a data guess is performed, so that there is room for the guesses.
		"""

		self.list.clear ()

	def DefaultCheck (self):
		"""
		Overview
		========
		This method has to deal with the preference (General Preferences) regarding where output of
		functions goes.  This setting is stored as a "global" in the bridge, so this method checks
		the bridge to see whether or not to send output to the parent basin object, or to the
		free variable viewer.
		"""

		if self._bridge._inPlace:
			self.pref.setCheckState (PyQt4.QtCore.Qt.Checked)
		else:
			self.pref.setCheckState (PyQt4.QtCore.Qt.Unchecked)

	def DoubleClicked (self, attr = None):
		"""
		Overview
		========
		This method will take an object and attempt to add it to the data selection of this dock widget.
		This can be slightly deceiving, as not just double clicking actions can invoke this
		method.  Things such as L{GuessAttributes ()<client.Dock.DataGuesser.GuessAttributes>} will
		also invoke it since it simulates a double clicking action.

		@param attr: This is the object from either the L{BasinViewer<client.Viewers.BasinViewer>}
		             or L{VariableList<client.Viewers.VariableViewer>} which was double clicked.
		@type attr: It can be one of many types
		             - L{Alias<client.Viewers.Alias>}
			     - L{Value<client.Viewers.Value>}
			     - L{Region<client.Viewers.Region>}
			     - L{List<client.Viewers.List>}
			     - L{Grid<client.Viewers.Grid>}
			     - L{Attribute<client.Viewers.Attribute>}
		"""

		if attr == None:
			return

		if attr.Type () == "ALIAS":
			attr = attr._item

		if attr.Type () == "ATTRIBUTE" or (attr.Type () == "VALUE" \
		and attr._type == "_basin.Attribute"):
			for holder in self.list._inputHolders:
				if holder.SetItem (attr, clobber = False) == True:
					self.DoubleClicked (holder._item._parent)
					return
		elif attr.Type () == "LIST":
			if self.list._parent.text () == "None":
				self.list._parent.setText (attr._name)
				self.list._parent._item = attr

	def setOutput (self, state):
		"""
		Overview
		========
		This method should be called whenever the "Transform in Place" checkbox's value changes.  This means
		it should be hooked up to the QCheckBox's I{void stateChanged (int)} signal.  When it is checked,
		the parent list is required for the data selection, so it is enabled.  Naturally, when the check
		box is unchecked, the parent list will be disabled.  Regardless of which state the check box is in,
		when this method is called the L{guessDefaults<client.Dock.Transforms.AbstractTransform.guessDefaults>}
		method is always called.

		@param state: The state is the state of the check box when its signal I{void stateChanged (int)}
		              is called.  We are only concerned with:
			       0. The check box was unchecked.
			       2. The check box was checked.
		@type state: int
		"""

		if   state == 0:
			self.list.setEnabled (False)
		elif state == 2:
			self.list.setEnabled (True)
		self.guessDefaults ()

	def transform (self):
		"""
		Overview
		========
		This method will attempt to execute the transformation on the data currently in the data selection.
		First it will make sure you are connected and already in autopx mode.  Then, depending on whether
		the user has selected if they want their data to be transformed in place or not, it will send
		the correct ipython command to the ipython shell for execution.  Afterwards, all of the data
		selection holders will be cleared.
		"""

		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False

		if   self.pref.checkState () == PyQt4.QtCore.Qt.Unchecked:
			inAttributes = ""
			result = self.list.getInAccesses ()
			if result == False:
				return
			for attr in result:
				inAttributes += ", " + str (attr)
			outAttributes = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttributes += ", " + str (attr)

			commands = []
			for out in self.list.getOutNames ():
				commands.append (str (out) + " = Attribute ()")


			commands.append (self._funcName + \
				" (" + str (inAttributes[2:]) + str (outAttributes) + ")")
			self._bridge._ipyShell.BatchCommands (commands, True)
		elif self.pref.checkState () == PyQt4.QtCore.Qt.Checked:
			inAttributes = ""
			result = self.list.getInNames ()
			if result == False:
				return
			for attr in result:
				inAttributes += ", \"" + str (attr) + "\""
			outAttributes = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttributes += ", \"" + str (attr) + "\""

			commands = []
			commands.append (self._funcName + \
				" (" + str (self.list._parent._item._shellAccess) + \
				str (inAttributes) + str (outAttributes) + ")")
			self._bridge._ipyShell.BatchCommands (commands, True)

		for holder in self.list._inputHolders:
			holder.clear ()
		for holder in self.list._outputHolders:
			holder.clear ()
		self.list.clear ()

	def guessDefaults (self):
		"""
		Overview
		========
		This function will first generate a list of regular expressions based on the
		input holders for the Transform Dock Widget.  From these regular expressions,
		it will run the Basin Object Tree items through the generated regular expressions,
		and on a match it will add it to the data selection layout.
		"""

		guessList = []
		for item in self._inList:
			guessList.append (DataGuesser.GenerateRegExpFromName (item))
		DataGuesser.GuessAttributes (self, guessList)


class Equatorial2XYZ (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "equatorial_to_xyz"
		self._displayName = "Equatorial To Cartesian"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/etoc.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["ra", "dec", "r"]
		self._outList = ["x", "y", "z"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class XYZ2Equatorial (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "xyz_to_equatorial"
		self._displayName = "Cartesian To Equatorial"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/ctoe.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["x", "y", "z"]
		self._outList = ["ra", "dec", "r"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class Polar2XY (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "polar_to_xy"
		self._displayName = "Polar To Cartesian"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/ptoc.png")
		self._description = "Everything this function does happens all within one list.  Choose two attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["r", "theta"]
		self._outList = ["x", "y"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class XY2Polar (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "xy_to_polar"
		self._displayName = "Cartesian To Polar"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/ctop.png")
		self._description = "Everything this function does happens all within one list.  Choose two attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!:"
		self._inList = ["x", "y"]
		self._outList = ["r", "theta"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class Spherical2XYZ (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "spherical_to_xyz"
		self._displayName = "Spherical To Cartesian"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/stoc.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["rho", "theta", "phi"]
		self._outList = ["x", "y", "z"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class XYZ2Spherical (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "xyz_to_spherical"
		self._displayName = "Cartesian To Spherical"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/ctos.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["x", "y", "z"]
		self._outList = ["rho", "theta", "phi"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class Cylindrical2XYZ (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "cylindrical_to_xyz"
		self._displayName = "Cylindrical To Cartesian"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/cytoc.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["r", "theta", "z"]
		self._outList = ["x", "y", "z"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class XYZ2Cylindrical (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._funcName = "xyz_to_cylindrical"
		self._displayName = "Cartesian To Cylindrical"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/ctocy.png")
		self._description = "Everything this function does happens all within one list.  Choose three attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["x", "y", "z"]
		self._outList = ["r", "theta", "z"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)


class Convolve (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._displayName = "Convolve"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/convolve.png")
		self._description = "This function convolves an attribute with a generic ConvKernel, which assumes a 2D Gaussian, which is 5x5. You may either convolve in the same List or Grid, or make it a free floating variable." 
		self._inList = ["in"]
		self._outList = ["conv_out"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)

	def guessDefaults (self):
		guessList = []
		DataGuesser.GuessAttributes (self, guessList)

	def transform (self):
		if not (self._bridge._connected == True and self._bridge._parallel == True):
			return False

		attributes = self.list.getInAccesses ()
		attr = ""
		if attributes != None and len (attributes) == 1:
			attr += attributes[0]
		else:
			return
		commands = []
		if   self.pref.checkState () == PyQt4.QtCore.Qt.Unchecked: # Free Floating
			commands.append (str (self.list._outList[0].text ()) + " = convolve (" + attr + ", CONVOLVE)")
		elif self.pref.checkState () == PyQt4.QtCore.Qt.Checked:   # In Place
			commands.append (str (self.list._item._shellAccess) + ".add_attribute (\"" + str (self.list._outList[0].text ()) + "\", convolve (" + attr + ", CONVOLVE))")
		self._bridge._ipyShell.BatchCommands (commands, True)

		for holder in self.list._inList:
			holder.clear ()
		for holder in self.list._outList:
			holder.clear ()
		self.list.clear ()


class FFT (AbstractTransform):
	def __init__ (self, parent, bridge):
		self._displayName = "Fast Fourier Transform"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/fft.png")
		self._description = "Everything this function does happens all within one list.  Choose one attributes from one list. These are the left-hand attributes.  You may name the right-hand ones, then click `Transform`!"
		self._inList = ["in"]
		self._outList = ["out"]
		# This constructor will make use of the previously defined variables
		# not defining them will cause an error
		AbstractTransform.__init__ (self, parent, bridge)

	def guessDefaults (self):
		return

	def transform (self):
		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False

		commands = []
		if   self.pref.checkState () == PyQt4.QtCore.Qt.Unchecked: # Free Floating
			commands.append (str (self.list._outList[0].text ()) + " = fft (" + str (self.list._inList[0]._item._shellAccess) + ")")
		elif self.pref.checkState () == PyQt4.QtCore.Qt.Checked:   # In Place
			commands.append (str (self.list._item._shellAccess) + ".add_attribute (\"" + str (self.list._outList[0].text ()) + "\", fft (" + str (self.list._item._shellAccess) + ".get_attribute (\"" + str (self.list._inList[0].text ()) + "\")))")
		self._bridge._ipyShell.BatchCommands (commands, True)

		for holder in self.list._inList:
			holder.clear ()
		for holder in self.list._outList:
			holder.clear ()
		self.list.clear ()


class ComputeDensity (AbstractTransform):
	"""
	"""

	def __init__ (self, layout, bridge):
		self._visitStarted = False
		self._displayName = "Compute Density"
		self._displayIcon = PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/transformations/compdens.png")
		self._description = "This function will take 1, 2, or 3 coordinates along with a mass to compute a density. The density output will be an attribute attached to a Grid called DensOut which is automatically added in place.  The output name you enter will be its alias."
		AbstractTransform.__init__ (self, layout, bridge)

	def _addOptions (self):
		return

	def _addCommands (self):
		box = PyQt4.QtGui.QGroupBox ("Commands")
		layout = PyQt4.QtGui.QHBoxLayout (box)

		commandTransform = Constants.BQPushButton ("Transform", layout)
		PyQt4.QtCore.QObject.connect (commandTransform, PyQt4.QtCore.SIGNAL ("clicked ()"), self.transform)

		commandClear = Constants.BQPushButton ("Clear", layout)
		PyQt4.QtCore.QObject.connect (commandClear, PyQt4.QtCore.SIGNAL ("clicked ()"), self.clear)

		return box

	def _addDataSelection (self):
		box = PyQt4.QtGui.QGroupBox ("Data Selection")
		layout = PyQt4.QtGui.QVBoxLayout (box)

		self._out = PyQt4.QtGui.QLineEdit ("densOut")
		outLayout = Constants.BQHBoxLayout (layout)
		outLayout.addWidget (PyQt4.QtGui.QLabel ("out:"))
		outLayout.addWidget (self._out)

		self._dataSelection =  Holders.ParentListHolder (self._bridge, ["x", "y", "z"])
		PyQt4.QtCore.QObject.connect (self._dataSelection._inList[0], PyQt4.QtCore.SIGNAL ("textChanged (const QString&)"), self.setNx)
		PyQt4.QtCore.QObject.connect (self._dataSelection._inList[1], PyQt4.QtCore.SIGNAL ("textChanged (const QString&)"), self.setNy)
		PyQt4.QtCore.QObject.connect (self._dataSelection._inList[2], PyQt4.QtCore.SIGNAL ("textChanged (const QString&)"), self.setNz)
		self._dataSelectionMassDens = Holders.FreeHolderGroup (self._bridge, ["mass"])


		self._nxName = PyQt4.QtGui.QLabel ("nx:")
		self._nx = PyQt4.QtGui.QLineEdit ("0")
		self._nx.setEnabled (False)
		self._nyName = PyQt4.QtGui.QLabel ("ny:")
		self._ny = PyQt4.QtGui.QLineEdit ("0")
		self._ny.setEnabled (False)
		self._nzName = PyQt4.QtGui.QLabel ("nz:")
		self._nz = PyQt4.QtGui.QLineEdit ("0")
		self._nz.setEnabled (False)

		self._dataSelection.layout.addWidget (self._nxName, 1, 2, PyQt4.QtCore.Qt.AlignRight)
		self._dataSelection.layout.addWidget (self._nx, 1, 3, PyQt4.QtCore.Qt.AlignLeft)
		self._dataSelection.layout.addWidget (self._nyName, 2, 2, PyQt4.QtCore.Qt.AlignRight)
		self._dataSelection.layout.addWidget (self._ny, 2, 3, PyQt4.QtCore.Qt.AlignLeft)
		self._dataSelection.layout.addWidget (self._nzName, 3, 2, PyQt4.QtCore.Qt.AlignRight)
		self._dataSelection.layout.addWidget (self._nz, 3, 3, PyQt4.QtCore.Qt.AlignLeft)

		space = PyQt4.QtGui.QHBoxLayout ()
		self._dataSelectionMassDens.layout.addLayout (space, 0, 2, 1, 2)

		regexp = PyQt4.QtCore.QRegExp ("[0-9]*")
		validatorX = PyQt4.QtGui.QRegExpValidator (regexp, self._nx)
		self._nx.setValidator (validatorX)
		validatorY = PyQt4.QtGui.QRegExpValidator (regexp, self._ny)
		self._ny.setValidator (validatorY)
		validatorZ = PyQt4.QtGui.QRegExpValidator (regexp, self._nz)
		self._nz.setValidator (validatorZ)

		layout.addLayout (self._dataSelection.layout)
		layout.addLayout (self._dataSelectionMassDens.layout)

		return box

	def setOutput (self):
		return

	def guessDefaults (self):
		guessList = []
		guessList.append (DataGuesser.GenerateRegExpFromName ("X"))
		guessList.append (DataGuesser.GenerateRegExpFromName ("Y"))
		guessList.append (DataGuesser.GenerateRegExpFromName ("Z"))
		guessList.append (DataGuesser.GenerateRegExpFromWord ("MASS"))
		DataGuesser.GuessAttributes (self, guessList)

	def transform (self):
		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False

		mass_dens = self._dataSelectionMassDens.getInNames (False)
		if len (mass_dens) == 1 and self._out.text () != "" \
		and self._dataSelection._item != None:
			beginning = "compute_density ( " + self._dataSelection._item._shellAccess + ", \""
			end = "\", \"" + mass_dens[0] + "\", \"" + str (self._out.text ())

			vectList = []
			for dim in [self._nx, self._ny, self._nz]:
				if dim.isEnabled () == True:
					vectList.append (str (dim.text ()))
			if   len (vectList) == 1:
				vect = "[" + vectList[0] + "]"
			elif len (vectList) == 2:
				vect = "[" + vectList[0] + ", " + vectList[1] + "]"
			elif len (vectList) == 3:
				vect = "[" + vectList[0] + ", " + vectList[1] + ", " + vectList[2] + "]"
			else:
				return

		else:
			return

		attributes = self._dataSelection.getInNames (False)


		if   len (attributes) == 3:
			self._bridge._ipyShell.BatchCommands ([beginning + \
				attributes[0] + "\", \"" + attributes[1] + "\", \"" + \
					attributes[2] +	end + "\", " + vect + ")"], True)
		elif len (attributes) == 2:
			self._bridge._ipyShell.BatchCommands ([beginning + \
				attributes[0] + "\", \"" + attributes[1] + \
				end + "\", " + vect + ", 2)"], True)
		elif len (attributes) == 1:
			self._bridge._ipyShell.BatchCommands ([beginning + \
				attributes[0] + end + "\", " + vect + ", 1)"], True)

		self.clear ()


	def setNx (self, val):
		if str (val) == "None":
			self._nx.setEnabled (False)
		else:
			self._nx.setEnabled (True)
			self._nx.setText ("50")

	def setNy (self, val):
		if str (val) == "None":
			self._ny.setEnabled (False)
		else:
			self._ny.setEnabled (True)
			self._ny.setText ("50")

	def setNz (self, val):
		if str (val) == "None":
			self._nz.setEnabled (False)
		else:
			self._nz.setEnabled (True)
			self._nz.setText ("50")

	def clear (self):
		self._dataSelection.clear ()
		for holder in self._dataSelection._inList:
			holder.clear ()
		for holder in self._dataSelectionMassDens._inList:
			holder.clear ()
		self._nx.setText ("0")
		self._ny.setText ("0")
		self._nz.setText ("0")
		self._out.setText ("densOut")

	def DoubleClicked (self, attr):
		if attr == None:
			return

		if attr.Type () == "ALIAS":
			attr = attr._item

		if attr.Type () == "ATTRIBUTE" or (attr.Type () == "VALUE" \
		and attr._type == "_basin.Attribute"):
			if self._dataSelection.isInFull ():
				if not self._dataSelectionMassDens.isInFull ():
					for holder in self._dataSelectionMassDens._inList:
						if holder.SetItem (attr, clobber = False) == True:
							return
					
			else:
				for holder in self._dataSelection._inList:
					if holder.SetItem (attr, clobber = False) == True:
						return
		elif attr.Type () == "LIST":
			if self._dataSelection.text () == "None":
				self._dataSelection.setText (attr.text (0))
				self._dataSelection._shellAccess = attr._shellAccess
