import PyQt4.QtGui
import PyQt4.QtCore

import string
import re

import Actions
import Constants

class Region (PyQt4.QtGui.QTreeWidgetItem):
	"""
	Overview
	========
	This object is the graphical representation of one sub-Region
	under the RegionTree.  For top-level Regions, a derived class
	TopRegion is used, which is mostly just named in a
	different manner.

	To constitute a valid Region, it must have a name and
	shellAccess.  The shellAccess is a string of the ipython command
	needed to access the object.  Since these are not variables,
	their names aren't their access.  Generally speaking, their
	shellAccess will be a traceback up the tree, with the appropriate
	calls.  The exception to this rule is with top-level regions
	where their name is indeed a shellAccess, where both object
	attributes still exist with the same value.

	Any abstract object, including this one, may have an alias.  An
	alias is nothing more than setting a variable to an Abstract object.
	The point being that some arbitrarily deep buried object might have
	an obscene shellAccess, but the user can assign it to a variable
	and the GUI will realize this and accomdate for it.

	These objects are created by having passed it a list of strings that
	identify the entire tree structure.  Rather than just the information
	specific to this tree branch, every node linearly "under" and including
	this node is listed.  It will find its current depth, and only generate
	tree items based on that depth (and tree structure "deeper"), until
	it has no more.  All mutations to the list of strings are permanent,
	so once a line generates a tree item, it disposes of itself.
	"""

	def __init__ (self, parent, parentShellAccess, info):
		"""
		"""

		self._parent = parent
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self)
		parent.addChild (self)

		background = PyQt4.QtGui.QColor (230, 230, 255)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(33, 0))
		gradient1.setColorAt (0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(100, 0))
		gradient2.setColorAt (0, background)
		gradient2.setColorAt (1, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackgroundColor (2, background)
		self.setBackground (3, brush2)
		self._alias = None
		currentDepth = info[0].count ("^")

		self.nameMe (parentShellAccess, info)
		info.pop (0)

		while len (info) > 0 and info[0].count ("^") > currentDepth:
			regexpList = re.compile ("\^*LIST:.*")
			regexpGrid = re.compile ("\^*GRID:.*")
			regexpChild = re.compile ("\^*REGION:.*")


			if   regexpList.match (info[0]):
				List (self, self._shellAccess, info)
			elif regexpGrid.match (info[0]):
				Grid (self, self._shellAccess, info)
			elif regexpChild.match (info[0]):
				Region (self, self._shellAccess, info)

		self.Display ()
		self.setExpanded (True)

	def Display (self):
		"""
		Overview
		========
		This function actually sets the text to screen. It will display the name, the fact
		that it is a Region, and if it exists, the alias name.  Notably, this is only 3
		columns where the RegionTree has 4 columns.  This is because no Regions have dimensions.
		"""

		self.setText (0, PyQt4.QtCore.QString (self._name))
		self.setText (1, PyQt4.QtCore.QString ("Region"))
		self.setTextAlignment (0, PyQt4.QtCore.Qt.AlignLeft | PyQt4.QtCore.Qt.AlignVCenter)
		self.setTextAlignment (1, PyQt4.QtCore.Qt.AlignRight | PyQt4.QtCore.Qt.AlignVCenter)
		if self._alias:
			self.setText (3, PyQt4.QtCore.QString (str (self._alias._name)))
		else:
			self.setText (3, PyQt4.QtCore.QString ("None"))

	def nameMe (self, parentShellAccess, info):
		"""
		@param parentShellAccess: Obviously the parent's shell access which directly contributes to the
					  child's shellAccess.
		@type parentShellAccess: string

		@param info:  This is the formatted list of strings that describes the entire makeup
			      of the tree.Information needed to determine the name is  extracted from
			      this list, and then destroyed from this list.
		@type info: list <string>
		"""

		if self.isTop () == False:
			regexp = re.compile ("\^*REGION:.*")
			if not regexp.match (info[0]):
				return
			self._name = info[0][info[0].index (":") + 1:]
			self._shellAccess = str (parentShellAccess) + ".get_child (\"" + self._name + "\")"

	def isTop (self):
		"""
		@return: For sub-Regions this will always return False.
		"""

		return False

	def Type (self):
		"""
		@return: For Regions, this will always return "REGION".
		"""
		return "REGION"

	def setAlias (self, alias):
		"""
		Overview
		========
		This method will take a given alias and set keep track of its reference to this Region. No checks are
		made currently, as this is assumed to be done in the calling function. This should only
		really be done from the L{VariableViewer<client.Viewers.VariableViewer>} when the alias has
		been established.

		@param alias: The alias which references this object.
		@type alias: L{Alias<client.Viewers.Items.Alias>}
		"""

		self._alias = alias
		self.Display ()

	def Delete (self):
		"""
		Overview
		========
		This method will create a dialog that simply verifies that the user wishes to delete this region.
		Note, this deletes the actual data object behind the abstract model. The data can't be recovered
		after performing this action.
		"""

		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # User clicked "Yes"
			result = self.treeWidget ()._bridge._ipyShell.Command (str (self.parent () \
				._shellAccess) + ".delete_child (" + self._shellAccess + ")", True)
			if result.find ("rror") != -1:
				self._bridge.Error ("BASIN_region_viewer.py", "Region",
					"Delete", result)
				return
			self.treeWidget ()._bridge.Update ()

	def Paste (self):
		"""
		Overview
		========
		This action assumes something has been copied.  Infact it only becomes available if something
		has been copied. The copied abstract object is stored on the RegionTree. Regions may have other
		Regions, Lists, or Grids under them.  If an attribute was copied, nothing will happen. After a
		Paste occurs, the copied object is no longer on the "clipboard".
		"""

		if   self.treeWidget ()._copy and self.treeWidget ()._copy.Type () == "REGION":
			if self.treeWidget ()._copy == self: return
			result = self.treeWidget ()._bridge._ipyShell.Command (str (self._shellAccess) +
				".add_child (" + self.treeWidget ()._copy._shellAccess + ")")
		elif self.treeWidget ()._copy and self.treeWidget ()._copy.Type () == "GRID":
			result = self.treeWidget ()._bridge._ipyShell.Command (str (self._shellAccess) +
				".add_grid (\"" + self.treeWidget ()._copy._name + "\", " +
				self.treeWidget ()._copy._shellAccess + ")")
		elif self.treeWidget ()._copy and self.treeWidget ()._copy.Type () == "LIST":
			result = self.treeWidget ()._bridge._ipyShell.Command (str (self._shellAccess) +
				".add_list (\"" + self.treeWidget ()._copy._name + "\", " +
				self.treeWidget ()._copy._shellAccess + ")")
		if result.find ("rror") != -1:
			self.treeWidget ()._bridge.Error ("BASIN_region_viewer.py", "Region", "Paste", result)

	def Copy (self):
		"""
		Overview
		========
		Just adds this object to the "clipboard", located on the RegionTree.  Doing so will allow Pastes
		to occur in the menu system.
		"""

		self.treeWidget ()._copy = self

	def CreateRegion (self):
		"""
		Overview
		========
		This will prompt the user for a name.  If a valid name is given, and the dialog wasn't cancelled,
		this method will create an empty Region as a child of the region being represented by this
		abstract region.
		"""

		item = self.treeWidget ().getSingleSelectedItem ()
		if item == None:
			return

		dialog = Actions.NameDialog (self.treeWidget ()._bridge._mainWindow, "Create Region...", PyQt4.QtGui.QIcon (Constants.ICON_DIR + \
			"/misc/createregion.png"))
		result = dialog.result ()
		if result == 1 and dialog.name.text () != "": # User clicked "Ok"
			shell = self.treeWidget ()._bridge._ipyShell
			commands = []
			shell.Command (str (dialog.name.text ()) + " = Region ()", False)

			shell.removeTopLevelRegion (str (dialog.name.text ()))
			commands.append (str (dialog.name.text ()) + ".set_name (\"" +
				str (dialog.name.text ()) + "\")")
			commands.append (str (item._shellAccess) + ".add_child (" +
				str (dialog.name.text ()) + ")")
			shell.BatchCommands (commands, True)

	def CreateGrid (self):
		"""
		Overview
		========
		This will prompt the user for a name.  If a valid name is given, and the dialog wasn't
		cancelled, this method will create an empty Grid as a child of the region being represented
		by this abstract region.
		"""

		item = self.treeWidget ().getSingleSelectedItem ()
		if item == None:
			return

		dialog = Actions.NameDialog (self.treeWidget ()._bridge._mainWindow, "Create Grid...", PyQt4.QtGui.QIcon (Constants.ICON_DIR + \
			"/misc/creategrid.png"))
		name = str (dialog.name.text ())
		result = dialog.result ()
		if result == 1 and name != "": # User clicked "Ok"
			commands = []
			commands.append (name + " = Grid ()")
			commands.append (str (item._shellAccess) + ".add_grid (\"" +
				str (dialog.name.text ()) + "\", " + name + ")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands)


	def CreateList (self):
		"""
		Overview
		========
		This will prompt the user for a name.  If a valid name is given, and the dialog wasn't
		cancelled, this method will create an empty List as a child of the region being represented
		by this abstract region.
		"""
		item = self.treeWidget ().getSingleSelectedItem ()
		if item == None:
			return
			
		dialog = Actions.NameDialog (self.treeWidget ()._bridge._mainWindow, "Create List...", PyQt4.QtGui.QIcon (Constants.ICON_DIR + \
			"/misc/createlist.png"))
		name = str (dialog.name.text ())
		result = dialog.result ()
		if result == 1 and name != "": # User clicked "Ok"
			commands = []
			commands.append (name + " = List ()")
			commands.append (str (item._shellAccess) + ".add_list (\"" +
				str (dialog.name.text ()) + "\", " + name + ")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, False)


class TopRegion (Region):
	"""
	Overview
	========
	This object is the graphical representation of one top-level Region under the RegionTree.
	To constitute a valid TopRegion, it must have a name and shellAccess.  The shellAccess is
	a string of the ipython command needed to access the object.  Their name is a shellAccess,
	where both object attributes still exist with the same value, since TopRegions are both
	basin objects and variables.

	Any abstract object, including this one, may have an alias.  An
	alias is nothing more than setting a variable to an Abstract object.
	The point being that some arbitrarily deep buried object might have
	an obscene shellAccess, but the user can assign it to a variable
	and the GUI will realize this and accomdate for it. Essentially 
	top-level Regions are aliases themselves, but you are welcome
	to give it another alias name if you wish.

	These objects are created by having passed it a list of strings that
	identify the entire tree structure.  Rather than just the information
	specific to this tree branch, every node linearly "under" and including
	this node is listed.  It will find its current depth, and only generate
	tree items based on that depth (and tree structure "deeper"), until
	it has no more.  All mutations to the list of strings are permanent,
	so once a line generates a tree item, it disposes of itself.
	"""
	def __init__ (self, parent, info, name):
		self._parent = parent
		self._name = name
		self._shellAccess = name
		Region.__init__ (self, parent, name, info)
		self.Display ()
			
	def isTop (self):
		"""
		TopRegion.isTop (self)

		For top-level Regions this will always return True.
		"""
		return True

	def Delete (self):
		"""
		TopRegion.Delete (self)

		This method will create a dialog that simply verifies
		that the user wishes to delete this region.  Note, this
		deletes the actual data object behind the abstract model.
		The data can't be recovered after performing this action.
		"""
		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # User clicked"Yes"
			self.treeWidget ()._bridge._ipyShell.removeTopLevelRegion (str (self._name))
			self.treeWidget ()._bridge._ipyShell.removeFreeVariable (str (self._name))
			commands = []
			commands.append ("del (" + self._shellAccess + ")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)
			self.treeWidget ()._bridge.Update ()


class Grid (PyQt4.QtGui.QTreeWidgetItem):
	def __init__ (self, parent, parentShellAccess, info):
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self)
		parent.addChild (self)
		self._parent = parent

		background = PyQt4.QtGui.QColor (230, 255, 230)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (33, 0))
		gradient1.setColorAt (0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (100, 0))
		gradient2.setColorAt (0, background)
		gradient2.setColorAt (1, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackgroundColor (2, background)
		self.setBackground (3, brush2)
		self._alias = None

		currentDepth = info[0].count ("^")
		self.nameMe (parentShellAccess, info)
		info.pop (0)

		while len (info) > 0 and info[0].count ("^") > currentDepth:
			regexp = re.compile ("\^*ATTRIBUTE:.*")
			if   regexp.match (info[0]):
				Attribute (self, self._shellAccess, info)

		self.Display ()
		self.setExpanded (True)

	def nameMe (self, parentShellAccess, info):
		regexp = re.compile ("\^*GRID:.*")
		if not regexp.match (info[0]):
			return
		self._name = info[0][info[0].index (":") + 1:]
		self._shellAccess = str (parentShellAccess) + ".get_grid (\"" + self._name + "\")"

	def Display (self):
		self.setText (0, PyQt4.QtCore.QString (str (self._name)))
		self.setText (1, PyQt4.QtCore.QString ("Grid"))
		self.setTextAlignment (0, PyQt4.QtCore.Qt.AlignLeft | PyQt4.QtCore.Qt.AlignVCenter)
		self.setTextAlignment (1, PyQt4.QtCore.Qt.AlignRight | PyQt4.QtCore.Qt.AlignVCenter)
		if self._alias:
			self.setText (3, PyQt4.QtCore.QString (str (self._alias._name)))
		else:
			self.setText (3, PyQt4.QtCore.QString ("None"))

	def Type (self):
		return "GRID"

	def setAlias (self, alias):
		self._alias = alias
		self.Display ()

	def Delete (self):
		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # User clicked "Yes"
			commands = []
			commands.append (str (self.parent () \
				._shellAccess) + ".delete_grid (\"" + self._name + "\")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)
			self.treeWidget ()._bridge.Update ()

	def Paste (self):
		if self.treeWidget ()._copy and self.treeWidget ()._copy.Type () == "ATTRIBUTE":
			commands = []
			commands.append (str (self._shellAccess) +
				".add_attribute (\"" + self.treeWidget ()._copy._name + "\", " +
				self.treeWidget ()._copy._shellAccess + ")")
			self.treeWidget ()._copy = None
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)

	def Copy (self):
		self.treeWidget ()._copy = self


class List (PyQt4.QtGui.QTreeWidgetItem):
	def __init__ (self, parent, parentShellAccess, info):
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self)
		parent.addChild (self)
		self._parent = parent

		background = PyQt4.QtGui.QColor (255, 230, 230)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (33, 0))
		gradient1.setColorAt (0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (100, 0))
		gradient2.setColorAt (0, background)
		gradient2.setColorAt (1, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackgroundColor (2, background)
		self.setBackground (3, brush2)
		self._alias = None

		currentDepth = info[0].count ("^")
		self.nameMe (parentShellAccess, info)
		info.pop (0)

		while len (info) > 0 and info[0].count ("^") > currentDepth:
			regexp = re.compile ("\^*ATTRIBUTE:.*")
			if   regexp.match (info[0]):
				Attribute (self, self._shellAccess, info)

		self.setExpanded (True)
		self.Display ()

	def nameMe (self, parentShellAccess, info):
		regexp = re.compile ("\^*LIST:.*")
		if not regexp.match (info[0]):
			return
		self._name = info[0][info[0].index (":") + 1:]
		self._shellAccess = str (parentShellAccess) + ".get_list (\"" + self._name + "\")"

	def Display (self):
		self.setText (0, PyQt4.QtCore.QString (str (self._name)))
		self.setText (1, PyQt4.QtCore.QString ("List"))
		self.setTextAlignment (0, PyQt4.QtCore.Qt.AlignLeft | PyQt4.QtCore.Qt.AlignVCenter)
		self.setTextAlignment (1, PyQt4.QtCore.Qt.AlignRight | PyQt4.QtCore.Qt.AlignVCenter)
		if self._alias:
			self.setText (3, PyQt4.QtCore.QString (str (self._alias._name)))
		else:
			self.setText (3, PyQt4.QtCore.QString ("None"))

	def Type (self):
		return "LIST"

	def setAlias (self, alias):
		self._alias = alias
		self.Display ()

	def Delete (self):
		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # User clicked "Yes"
			commands = []
			commands.append  (str (self.parent ()._shellAccess) +
				".delete_list (\"" + self._name + "\")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)
			self.treeWidget ()._bridge.Update ()

	def Paste (self):
		if self.treeWidget ()._copy and self.treeWidget ()._copy.Type () == "ATTRIBUTE":
			commands = []
			commands.append (str (self._shellAccess) +
				".add_attribute (\"" + self.treeWidget ()._copy._name + "\", " +
				self.treeWidget ()._copy._shellAccess + ")")
			self.treeWidget ()._copy = None
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)

	def Copy (self):
		self.treeWidget ()._copy = self


class Attribute (PyQt4.QtGui.QTreeWidgetItem):
	# Standard init that will add it to the parent item,
	# after probing ipython for its particular information
	def __init__ (self, parent, parentShellAccess, info):
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self)
		parent.addChild (self)
		self._parent = parent

		background = PyQt4.QtGui.QColor (230, 230, 230)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(33, 0))
		gradient1.setColorAt (0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(100, 0))
		gradient2.setColorAt (0, background)
		gradient2.setColorAt (1, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackgroundColor (2, background)
		self.setBackground (3, brush2)
		self._alias = None

		currentDepth = info[0].count ("^")
		self.nameMe (parentShellAccess, info)
		info.pop (0)

		self.getDims (info)
		info.pop (0)

		self.Display ()

	# Display information for the item in the QTreeWidget...
	# It should display the name, the fact that its an attribute,
	# its dimensionality, and its alias, if any
	def Display (self):
		self.setText (0, PyQt4.QtCore.QString (str (self._name)))
		self.setText (1, PyQt4.QtCore.QString ("Attribute"))
		self.setText (2, PyQt4.QtCore.QString (str (self._dims)))
		if self._alias:
			self.setText (3, PyQt4.QtCore.QString (str (self._alias._name)))
		else:
			self.setText (3, PyQt4.QtCore.QString ("None"))
		self.setTextAlignment (0, PyQt4.QtCore.Qt.AlignLeft | PyQt4.QtCore.Qt.AlignVCenter)
		self.setTextAlignment (1, PyQt4.QtCore.Qt.AlignRight | PyQt4.QtCore.Qt.AlignVCenter)
		self.setTextAlignment (2, PyQt4.QtCore.Qt.AlignRight | PyQt4.QtCore.Qt.AlignVCenter)

	# retrieves the dimensionality of an attribute via probing the ipython session
	# It should be mentioned, in basin an attribute's name appears twice, one in the attribute, and also
	# in the list/grid's map that maps to the attribute
	def nameMe (self, parentShellAccess, info):
		regexp = re.compile ("\^*ATTRIBUTE:.*")
		if not regexp.match (info[0]):
			return
		self._name = info[0][info[0].index (":") + 1:]
		self._shellAccess = str (parentShellAccess) + ".get_attribute (\"" + self._name + "\")"

	# retrieves the dimensionality of an attribute via probing the ipython session
	def getDims (self, info):
		regexp = re.compile ("\^*DIMENSION:.*")
		if not regexp.match (info[0]):
			return
		self._dims = info[0][info[0].index (":") + 1:]

	# Returns the type, this is useful because often this is mixed in
	# which objects of types LIST, GRID, or REGION
	# and specific ones need to be targeted
	def Type (self):
		return "ATTRIBUTE"

	# Sets the alias, which is a PyQt4.QtGui.QTreeWidgetItem from the variableViewer
	# This is also called from the variableViewer, when aliases find their reference, they
	# call setAlias on their reference
	def setAlias (self, alias):
		self._alias = alias
		self.Display ()

	# Calls the class Actions.DeleteDialog, and if it is successful deletes it from its parent
	def Delete (self):
		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # User clicked "Yes"
			# The only way to delete an attribute is from its parent,
			# and with the string name of the attribute you wish to delete.
			commands = []
			commands.append (str (self.parent ()._shellAccess) +
				".delete_attribute (\"" + self._name + "\")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)
			self.treeWidget ()._bridge.Update ()

	def Copy (self):
		self.treeWidget ()._copy = self


class Alias (PyQt4.QtGui.QTreeWidgetItem):
	def __init__ (self, parent, name, access):
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self, parent)

		background = PyQt4.QtGui.QColor (255, 228, 181)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(33, 0))
		gradient1.setColorAt (0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF(100, 0))
		gradient2.setColorAt (0, background)
		gradient2.setColorAt (1, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackground (2, brush2)
		self.setText (0, PyQt4.QtCore.QString (str (name)))
		self.setText (1, PyQt4.QtCore.QString ("Alias"))
		self._shellAccess = str (access)
		self._item = None
		self._name = name
		self.findReference ()

	def Type (self):
		return "ALIAS"

	def findReference (self):
		iter = PyQt4.QtGui.QTreeWidgetItemIterator (self.treeWidget ()._bridge._regionViewer)
		while iter.value ():
			if iter.value ().text (0) != "BASIN":
				if self._shellAccess.replace (" ", "") == \
				iter.value ()._shellAccess.replace (" ", ""):
					self.setText (2, PyQt4.QtCore.QString (str (iter.value ()._name)))
					iter.value ().setAlias (self)
					self._item = iter.value ()
					return True
			iter += 1
		return False

	def Delete (self):
		if self._item.Type () == "REGION" and self._item.isTop ():
			# Deleting the aliased item will also delete this
			self._item.Delete ()
		else:
			self.treeWidget ()._bridge._ipyShell.removeFreeVariable (self._name)
			self.treeWidget ()._bridge.Update ()


class Value (PyQt4.QtGui.QTreeWidgetItem):
	def __init__ (self, parent, name):
		PyQt4.QtGui.QTreeWidgetItem.__init__ (self, parent)

		background = PyQt4.QtGui.QColor (152, 251, 152)
		gradient1 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (33, 0))
		gradient1.setColorAt (0.0, PyQt4.QtCore.Qt.white)
		gradient1.setColorAt (1.0, background)
		brush1 = PyQt4.QtGui.QBrush (gradient1)
		gradient2 = PyQt4.QtGui.QLinearGradient (PyQt4.QtCore.QPointF(0, 0), PyQt4.QtCore.QPointF (100, 0))
		gradient2.setColorAt (0.0, background)
		gradient2.setColorAt (1.0, PyQt4.QtCore.Qt.white)
		brush2 = PyQt4.QtGui.QBrush (gradient2)

		self.setBackground (0, brush1)
		self.setBackgroundColor (1, background)
		self.setBackground (2, brush2)
		self.setText (0, PyQt4.QtCore.QString (str (name)))

		# where everything else usually references something else, a value is itself, and so
		# these values are redundant, but make it look the same as any other object
		self._item = self
		self._name = name
		self._shellAccess = name
		self._type = self.getRealType ()
		self.setText (1, PyQt4.QtCore.QString (str (self._type)))

	def Type (self):
		return "VALUE"

	def getRealType (self):
		result = self.treeWidget ()._bridge._ipyShell.ControllerCommand \
			("print str (type (" + self._name + "))")
		if result.find ("Error") != -1:
			self.treeWidget ()._bridge.Error ("BASIN_variable_viewer.py", \
				"VariableValue", "getRealType", result)
			return "#Error"
		begin = result.find ("'")
		close = result.rfind ("'")
		return result[begin + 1:close]

	def Delete (self):
		dialog = Actions.DeleteDialog (self.treeWidget ()._bridge._mainWindow, self._name)
		result = dialog.result ()
		if result == 1: # "Yes"
			self.treeWidget ()._bridge._ipyShell.removeFreeVariable (str (self._name))
			commands = []
			commands.append ("del (" + self._shellAccess + ")")
			self.treeWidget ()._bridge._ipyShell.BatchCommands (commands, True)
			self.treeWidget ()._bridge.Update ()

