from PyQt4.QtGui import *
from PyQt4.QtCore import *

import re

import Items
import Actions

class BasinViewer (QTreeWidget):
	"""
	Overview
	========
	This class holds a graphical representation (in a tree format) of all the BASIN objects
	currently accounted for by the GUI. It will only keep track of those BASIN objects that
	were made while in %autopx mode (IPython1's auto-parallel mode).

	The general way that it does this is by relying on the BASIN_ipython_shell.IPythonView
	object for a few things. First, the IPythonView must be able to catch each time a top-level
	region is created.  Then it must stored that information on the parallel engine by appending
	the name of this new top-level region to a variable named BASIN_TOPLEVEL_REGION_LIST.  It
	is a newline deliminated string containing the names of top-level Regions.

	Since Regions are able to self-examine themselves, all
	that needs to be stored is their variable name.  From
	that it can get the total information about objects
	that are children to it.  From their it constructs the
	tree using the following classes:
	 - Region
	 - Grid
	 - List
	 - Attribute
	"""

	def __init__ (self, layout, parent, bridge):
		"""

		This sets up the QTreeWidget to have 4 columns:
		 1. Data       - The name of the data
		 2. Type       - It's BASIN object type
		 3. Dimensions - Applicable only to BASIN Attributes
		 4. Alias      - The name of the alias, if it exists
		                 Aliases are actual python variables
			         that reference the represented

		@param layout: The parent's layout.  This will be automatically added to the layout.
		@type layout: 

		@param parent: The parent object of this tree.
		@type parent: 

		@param bridge: The overall connecting bridge class which gets set to self._bridge
		@type bridge: L{Bridge<client.BASIN_bridge.Bridge>}
		"""

		QTreeWidget.__init__ (self, parent)

		self._bridge = bridge
		bridge.setRegionViewer (self)

		self.setAnimated (False)
		headers = QStringList ("Data")
		headers.append ("Type")
		headers.append ("Dimensions")
		headers.append ("Alias")
		self.setHeaderLabels (headers)
		self.setDragEnabled (True)
		self.setAcceptDrops (True)
		self._copy = None

		self.makeBasin ()

		box = QGroupBox ("BASIN Data Structure")
		boxLayout = QHBoxLayout (box)
		boxLayout.addWidget (self)
		layout.addWidget (box)
	
	def Update (self, topRegions):
		"""
		Overview
		========
		This should only ever be called from the BASIN_bridge.Bridge class.

		It clears out all current entries and then re-populates them based off of a (hopefully)
		new BASIN_TOPLEVEL_REGION_LIST.

		Presumably this program also uses a BASIN_variable_viewer.VariableList object.
		In such a case, this must be updated BEFORE the VariableList is, and thus it should
		really only be handled at one point, thus I let the bridge handle Update.
		"""

		self.clear ()
		self._copy = None
		self.makeBasin ()

		for region in topRegions:
			self.MakeTopRegion (region)

		self.expandItem (self.top)
		self.resizeColumnToContents (2)
		self.resizeColumnToContents (1)
		self.resizeColumnToContents (0)

	def getSingleSelectedItem (self):
		"""
		Overview
		========
		This will retrieve the item currently selected by the user.  It is possible (although unlikely)
		that more than one, or none, may be selected. In such a case, a False signal is returned.
		"""

		items = self.selectedItems ()
		if len (items) == 1:
			return items[0]

	def makeBasin (self):
		"""
		Overview
		========
		This is called initially, and after each Update. All it doesi s create the top-most object in the
		tree structure.  It has no real value, just a place-holder for everything else to go under, much
		like a 'root' that must exist first, regardless of anything.
		"""

		self.top = QTreeWidgetItem ()
		self.top.setText (0, "BASIN")
		self.top.setFlags (Qt.ItemIsDropEnabled)
		self.addTopLevelItem (self.top)

	def MakeTopRegion (self, var):
		"""
		Overview
		========
		This will be called by Update for as many times as there are top-level Regions.  It will take the
		name of said Region and call its `basin_kernel` method `Region::get_client_information`.  This
		generates a string containing information about all sub-objects of this Region.

		All This really	does is take the name of the Region, and this information string and pass
		it to the constructor of BASIN_region_viewer.TopRegion along with the reference to
		the top-most root in the tree, where new top-level regions will be under.
		"""

		realResult = []
		result = self._bridge._ipyShell.SilentCommand ("print \"\\n\" + " + var +
			".get_client_information ()").split ("\n")[3:]

		# .index () will result in an error if the argument doesn't exist,
		# so use count to ensure no errors occur
		if result.count ("") > 0:
			result = result[:result.index ("")]

		regexp = re.compile (".*\[.*\].*(In|Out).*\[.*\]")
		for line in result:
			if not regexp.match (line):
				realResult.append (line)

		Items.TopRegion (self.top, realResult, var)

	def mouseMoveEvent (self, ev):
		"""
		Overview
		========
		This is used to initiate a drag sequence. Basically a drag will only start when the user
		has pressed on an object, and moved a minimum distance.  This is to prevent slight movements
		during double-clicks as being interepreted incorrectly.
		"""

		if (ev.globalPos () - self._dragStartPosition).manhattanLength () > \
			QApplication.startDragDistance ():
			object = self.getSingleSelectedItem ()
			if object != None:
				mime = QMimeData ()
				mime.setData (QString ("text/plain"), str (object._name + "\n" +
					object._shellAccess + "\n" + object.Type ()))
				drag = QDrag (self)
				drag.setMimeData (mime)
				self._bridge._drugItem = object
				dropAction = drag.start (Qt.LinkAction)

     	def mousePressEvent (self, ev):
		"""
		Overview
		========
		Captures the initial position of a mouse button being pressed.  The onyl real use for
		this is in conjunction with mouseMoveEvent, which will initiate a drag sequence if the
		mouse has moved a minimum distance from this initial point.
		"""

		QTreeWidget.mousePressEvent (self, ev)
		self._dragStartPosition = ev.pos ()
		item = self.getSingleSelectedItem ()
		if item and item.Type () == "REGION":
			self._bridge._actions.menuRegionSelected ()
		else:
			self._bridge._actions.menuRegionNotSelected ()

	def mouseDoubleClickEvent (self, ev):
		"""
		Overview
		========
		This method merely grabs the currently selected item and sends it on its way to the
		BASIN_dock.Dock.  Presumably the Dock can handle it from there.
		"""

		self._bridge.AddToDock (self.getSingleSelectedItem ())

	def mouseReleaseEvent (self, ev):
		"""
		Overview
		========
		This method will check to see if the user released the right button of the mouse while
		over the RegionTree.  If so, it sends the position of the release to the appropriate
		handling method, rightClick.
		"""

		QTreeWidget.mouseReleaseEvent (self, ev)
		if ev.button () == Qt.RightButton:
			self.rightClick (ev.globalPos ())

	def rightClick (self, position):
		"""
		Overview
		========
		This method is obviously called on a right-click and drops down a menu generated by this function.

		It depends on the exact object you clicked on, for what options will appear in the menu.

		Generally speaking the user will see options for:
		 - Copying / Pasting (All BASIN types)
		 - Creating an Alias (All BASIN types)
		 - Deleting (All BASIN types)
		 - Creating an appropriate blank sub-object
		 - Exporting (Regions only)
		 - Adding it to the Dock Widget (All BASIN types,
		   though there is no garauntee the dock will do
		   anything useful with it)
		"""

		if (not self._bridge._connected) or (not self._bridge._parallel):
			return
		item = self.getSingleSelectedItem ()
		if item == None:
			menu = QMenu ("Right-Click", self._bridge._mainWindow)
			menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/createregion.png")), QString ("Create Region"),
				self.CreateRegion)
			menu.exec_ (position)
			return

		menu = QMenu ("Right-Click", self._bridge._mainWindow)
		menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/copy.png")), QString ("Copy"), item.Copy)
		if item.Type () != "ATTRIBUTE":
			paste = menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/paste.png")), QString ("Paste"), item.Paste)
			if not self._copy:
				paste.setEnabled (False)
			else:
				if   item.Type () == "REGION" and  self._copy.Type () == "ATTRIBUTE":
					paste.setEnabled (False)
				elif (item.Type () == "GRID" or item.Type () == "LIST") and \
					not self._copy.Type () == "ATTRIBUTE":
					paste.setEnabled (False)
		menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/delete.png")), QString ("Delete"), item.Delete)

		if   item.Type () == "REGION":
			menu.addSeparator ()
			menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/createregion.png")),
				QString ("Create Region"),
				item.CreateRegion)
			menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/creategrid.png")),
				QString ("Create Grid"),
				item.CreateGrid)
			menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/createlist.png")),
				QString ("Create List"),
				item.CreateList)

		menu.addSeparator ()

		menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/alias.png")), QString ("Create Alias..."),
			self._bridge._varViewer.createAlias)

		if   item.Type () == "ATTRIBUTE":
			menu.addAction ("Convert to list...", self.AttributeToList)

		menu.exec_ (position)

	def AttributeToList (self):
		"""
		"""

		item = self.getSingleSelectedItem ()
		if item == None:
			return

		dialog = Actions.NameDialog (self._bridge._mainWindow, "Convert To List...", QIcon (Constants.ICON_DEFAULT))
		result = dialog.result ()
		if result == 1 and dialog.name.text () != "": # "Ok"
			self._bridge._ipyShell.Command (str (dialog.name.text ()) + " = " + \
				str (item._shellAccess) + ".to_list ()", True)

	def CreateRegion (self):
		"""
		Overview
		========
		This method will potentially create a top-level Region.  It will call a dialog which will request
		a name from the user, or they can cancel.  The given name will become the name of the new top-
		level Region, or nothing will happen given a blank name / cancel.
		"""

		item = self.getSingleSelectedItem ()

		if item == None:
			dialog = NameDialog (self._bridge._mainWindow, "Top Region", QIcon (Constants.ICON_DIR + \
				"/misc/createregion.png"))
			result = dialog.result ()
			if result == 1 and dialog.name.text () != "": # "Ok"
				commands = []
				commands.append  (str (dialog.name.text ()) + " = Region ()")
				commands.append (str (dialog.name.text ()) +
					".set_name (\"" + str (dialog.name.text ()) + "\")")
				self._bridge._ipyShell.BatchCommands (commands, True)

     	def dragEnterEvent (self, ev):
		"""
		Overview
		========
		This initial drag event merely checks to make sure that the source of the drag came from the
		BASIN_variable_viewer.List object.
		"""

		if ev.source () == self._bridge._varViewer:
			ev.accept ()
		else:
			ev.ignore ()

	def dragMoveEvent (self, ev):
		"""
		Overview
		========
		This midway drag  move event merely checks to make sure that the source of the drag came from the
		BASIN_variable_viewer.List object.
		"""

		if ev.source () == self._bridge._varViewer:
			ev.accept ()
		else:
			ev.ignore ()

	def dropEvent (self, ev):
		"""
		Overview
		========
		This method will determine the type of free variable being dropped on the RegionTree, and
		the exact object type it was dropped on.
		"""

		if ev.source () == self._bridge._varViewer:
			# The source is the free variable
			source = self._bridge.PopDrugItem ()
			# And the target is the region-list structure they drop it on
			target = self.itemAt (ev.pos ())

			# There's no certainty that they actually dropped it on something...
			if target == None:
				ev.ignore ()
				return

			if source.Type () != "VALUE" and source._alias != None:
				self._bridge._statusBar.showTimed ("Can't add an alias, it is already " + \
					"coupled to another item!", 2.5)
				return

			if   source.Type () == "ATTRIBUTE" or (source.Type () == "VALUE" and \
				source._type == "_basin.Attribute"):
				if target.Type () == "LIST" or target.Type () == "GRID":
					self._appendAttributeFromFreeVar (source._shellAccess, target._shellAccess)
				else:
					self._bridge._statusBar.showTimed ("Attributes may only be appended to Lists or Grids", 2.5)
			elif source.Type () == "GRID" or (source.Type () == "VALUE" and \
				source._type == "_basin.Grid"):
				if target.Type () == "REGION":
					self._appendGridFromFreeVar (source._shellAccess, target._shellAccess)
				else:
					self._bridge._statusBar.showTimed ("Grids may only be appended to Regions", 2.5)
			elif source.Type () == "LIST" or (source.Type () == "VALUE" and \
				source._type == "_basin.List"):
				if target.Type () == "REGION":
					self._appendListFromFreeVar (source._shellAccess, target._shellAccess)
				else:
					self._bridge._statusBar.showTimed ("Lists may only be appended to Regions", 2.5)
			elif source.Type () == "REGION" or (source.Type () == "VALUE" and \
				source._type == "_basin.Region"):
				if target.Type () == "REGION":
					self._appendRegionFromFreeVar (source._shellAccess, target._shellAccess)
				else:
					self._bridge._statusBar.showTimed ("Regions may only be appended to Regions or BASIN", 2.5)
			else:
				self._bridge._statusBar.showTimed ("Only BASIN objects may be appended to the Region Viewer", 2.5)

			ev.setDropAction (Qt.LinkAction)
			ev.accept ()
		else:
			ev.ignore ()

	def _appendAttributeFromFreeVar (self, source, target):
		"""
		Overview
		========
		This will occur when a user drags a free variable BASIN Attribute and drops it onto
		a RegionTree object that is capable of being its parent, in this case Lists or Grids.

		To accomplish this, it can use the internal `basin_kernel` command to add the attribute.
		Also, the GUI will need to be updated that this has made.  This is because we aren't
		destroying the free floating attribute, but it just turns into an alias for the object
		you are adGui ding into the RegionTree.
		"""

		commands = []
		commands.append (target + ".add_attribute (\"" + source +
			"\", " + source + ")")
		self._bridge._ipyShell.replaceFreeVariable (source, target + ".get_attribute (\"" +
			source + "\")")
		self._bridge._ipyShell.BatchCommands (commands, True)

	def _appendGridFromFreeVar (self, source, target):
		"""
		Overview
		========
		This will occur when a user drags a free variable BASIN Grid and drops it onto a
		RegionTree object that is capable of being its parent, in this case Regions.

		To accomplish this, it can use the internal `basin_kernel` command to add the Grid.
		Also, the GUI will need to be updated that this has made.  This is because we aren't
		destroying the free floating Grid, but it just turns into an alias for the object
		you are adding into the RegionTree.
		"""

		commands = []
		commands.append (target + ".add_grid (\"" + source + "\", " +
			source + ")")
		self._bridge._ipyShell.replaceFreeVariable (source, target + ".get_grid (\"" +
			source + "\")")
		self._bridge._ipyShell.BatchCommands (commands, True)

	def _appendListFromFreeVar (self, source, target):
		"""
		Overview
		========
		This will occur when a user drags a free variable BASIN List and drops it onto a
		RegionTree object that is capable of being its parent, in this case Regions.

		To accomplish this, it can use the internal `basin_kernel` command to add the List.
		Also, the GUI will need to be updated that this has made.  This is because we aren't
		destroying the free floating List, but it just turns into an alias for the object you
		are adding into the RegionTree.
		"""

		commands = []
		commands.append (target + ".add_list (\"" + source +
			"\", " + source + ")")
		self._bridge._ipyShell.replaceFreeVariable (source, target + ".get_list (\"" +
			source + "\")")
		self._bridge._ipyShell.BatchCommands (commands, True)

	def _appendRegionFromFreeVar (self, source, target):
		"""
		Overview
		========
		This will occur when a user drags a free variable BASIN Region and drops it onto a
		RegionTree object that is capable of being its parent, in this case Regions.

		To accomplish this, it can use the internal `basin_kernel` command to add the Region.
		Also, the GUI will need to be updated that this has made.  This is because we aren't
		destroying the free floating Region, but it just turns into an alias for the object
		you are adding into the RegionTree.
		"""

		commands = []
		commands.append (target + ".add_child (" + source + ")")
		self._bridge._ipyShell.BatchCommands (commands, True)


class VariableViewer (QTreeWidget):
	"""
	Overview
	========
	This class graphically keeps track of all free floating variables within the python session.
	Note, it keeps track of variables in the parallel session, or commands entered while in
	%autopx mode.  There are essentially two types of free floating variables that are represented:
	 - Aliases: Variables with point to structured Basin objects.  Note that not all basin objects
	            are aliases.  An attribute without a list, for instance, isn't structured.
	 - Values: Self contained variables that are not references to a Basin object.
	"""
	def __init__ (self, layout, bridge):
		"""
		"""
		QTreeWidget.__init__ (self)
		self.setAcceptDrops (True)

		self._bridge = bridge
		bridge.setVariableViewer (self)

		headers = QStringList ("Variables")
		headers.append ("Type")
		headers.append ("Reference")
		self.setHeaderLabels (headers)
		self.setDragEnabled (True)
		self.setAcceptDrops (True)

		self.makePython ()

		box = QGroupBox ("Free Python Variables")
		boxLayout = QHBoxLayout (box)
		boxLayout.addWidget (self)
		layout.addWidget (box)

	def Update (self, freeInfo):
		freeVars = freeInfo[0]
		freeAccess = freeInfo[1]
		self.clear ()
		self.makePython ()
		if len (freeVars) == 0 or len (freeAccess) == 0:
			return
		for index in xrange (len (freeVars)):
			if freeVars.count (freeVars[index]) < 2:
				result = self.findReference (freeAccess[index])
				if result == False:
					Items.Value (self.top, freeVars[index])
				else:
					Items.Alias (self.top, str (freeVars[index]), str (freeAccess[index]))
		self.expandItem (self.top)
		self.resizeColumnToContents (2)
		self.resizeColumnToContents (1)
		self.resizeColumnToContents (0)

	def makePython (self):
		self.top = QTreeWidgetItem ()
		self.top.setText (0, "Python")
		self.top.setFlags (Qt.ItemIsDropEnabled)
		self.addTopLevelItem (self.top)

	def isBasinObject (self, name):
		result =  self._bridge._ipyShell.SilentCommand ("print \"\\n\" + str (" + name + ")")
		if result.find ("Error") != -1:
			self._bridge.Error ("BASIN_variable_viewer.py", "VariableViewer",
				"isBasinObject", result)
			return False
		result = result.split ("\n")[3]
		if re.compile ("<_basin.*>").match (result):
			return True
		return False

	def findReference (self, shellAccess):
		iter = QTreeWidgetItemIterator (self._bridge._regionViewer)
		iter += 1
		while iter.value ():
			if shellAccess.replace (" ", "") == \
			iter.value ()._shellAccess.replace (" ", ""):
				return True
			iter += 1
		return False

	def createAlias (self, attr = None):
		if attr == None:
			attr = self._bridge._regionViewer.getSingleSelectedItem ()
			if attr == None:
				return
		dialog = QDialog ()
		dialog.setWindowTitle (QString ("Create Alias..."))
		dialog.setWindowIcon (QIcon (QString (Constants.ICON_DIR + "/alias.png")))
		layout = QVBoxLayout (dialog)
		topLayout = Constants.BQHBoxLayout (layout)
		bottomLayout = Constants.BQHBoxLayout (layout)

		name = QLineEdit (attr._name)
		regexp = QRegExp ("[A-Za-z_][0-9A-Za-z_]*")
		valid = QRegExpValidator (regexp, name)
		name.setValidator (valid)
		topLayout.addWidget (QLabel (QString ("Alias Name:")))
		topLayout.addWidget (name)

		acceptButton = Constants.BQPushButton ("Ok", bottomLayout)
		QObject.connect (acceptButton, SIGNAL ("clicked ()"), dialog.accept)
		rejectButton = Constants.BQPushButton ("Cancel", bottomLayout)
		QObject.connect (rejectButton, SIGNAL ("clicked ()"), dialog.reject)

		self._bridge.CursorNorm ()
		result = dialog.exec_ ()

		if result == 1: # "Ok"
			result = self._bridge._ipyShell.Command (str (name.text ()) + " = " +
				str (attr._shellAccess))
			if result.find ("Error") != -1:
				self._bridge.Error ("BASIN_variable_viewer.py", "VariableViewer",
					"createAlias", result)

	def getSingleSelectedItem (self):
		items = self.selectedItems ()
		if len (items) == 1:
			return items[0]
		return None

	def dropEvent (self, ev):
		if ev.source () == self._bridge._regionViewer:
			self.createAlias (self._bridge.PopDrugItem ())

			ev.setDropAction (Qt.LinkAction)
			ev.accept ()
		else:
			ev.ignore ()

     	def dragEnterEvent (self, ev):
		if ev.source () == self._bridge._regionViewer:
			ev.accept ()
		else:
			ev.ignore ()

	def dragMoveEvent (self, ev):
		if ev.source () == self._bridge._regionViewer:
			ev.accept ()
		else:
			ev.ignore ()

	def mouseDoubleClickEvent (self, ev):
		self._bridge.AddToDock (self.getSingleSelectedItem ())
		QTreeWidget.mouseDoubleClickEvent (self, ev)

	def mouseMoveEvent (self, ev):
		if (ev.globalPos () - self._dragStartPosition).manhattanLength () > \
		QApplication.startDragDistance ():
			object = self.getSingleSelectedItem ()
			if object != None:
				mime = QMimeData ()
				mime.setData (QString ("text/plain"), "")
				drag = QDrag (self)
				drag.setMimeData (mime)
				if   object.Type () == "ALIAS":
					self._bridge._drugItem = object
					#self._bridge._drugItem = object._item
				elif object.Type () == "VALUE":
					self._bridge._drugItem = object
				dropAction = drag.start (Qt.LinkAction)

     	def mousePressEvent (self, ev):
		QTreeWidget.mousePressEvent (self, ev)
		self._dragStartPosition = ev.pos ()
		item = self.getSingleSelectedItem ()

	def mouseReleaseEvent (self, ev):
		QTreeWidget.mouseReleaseEvent (self, ev)
		if ev.button () == Qt.RightButton:
			self.rightClick (ev.globalPos ())

	def rightClick (self, position):
		item = self.getSingleSelectedItem ()
		if item == None:
			return

		menu = QMenu ("Right-Click", self._bridge._mainWindow)
		if item.Type () == "ALIAS":
			if item.Type () == "REGION":
				menu.addAction (self._bridge._actions.fExport)
		elif item.Type () == "VALUE":
			item = item

		menu.addAction (QIcon (QString (Constants.ICON_DIR + "/misc/delete.png")), QString ("Delete"), item.Delete)

		menu.exec_ (position)
